from typing import Union, Tuple, Any, List, Type, Dict
from dataclasses import dataclass
from pathlib import Path
from collections import OrderedDict
from enum import Enum

from . import ModelFormat, ModelDataType
from ..common import get_class_name




@dataclass
class Tensor:
    name: str
    shape: Union[None, Tuple[int, ...]] = None
    dtype: Union[None, ModelDataType] = None

    def dump(self) -> Dict:
        data = {
            "name": self.name,
            "shape": self.shape,
            "type": self.dtype.name
        }
        return data


class Model():
    def __init__(self, src: Union[str, Path, object]):
        self.src = src
        self.inputs: List[Tensor] = []
        self.outputs: List = []
        self.format: ModelFormat = ModelFormat.NON_SPECIFIED
        self._name: str

        if isinstance(src, str):
            self._name = Path(src).stem
        elif isinstance(src, Path):
            self._name = src.stem
        else:
            self._name = str(type(object))

    @property
    def name(self) -> str:
        return self._name

    def set_name(self, _name: str):
        self._name = _name

    def add_input(self, input_: Tensor) -> None:
        self.inputs.append(input_)

    def add_output(self, output_: Tensor) -> None:
        self.outputs.append(output_)

    def clone_attributes(self, model: Any) -> None:
        """Clone user defined attributes from input

        Iter all user-defined attributes and check if they have values.
        If the attribute of the destination object has value, then skip
        the clone.

        Args:
            model (Type[Dataset]): The model object to be cloned
        """
        _to_be_cloned: List = []
        for attr_name in model.__dict__:

            dst_value = getattr(self, attr_name)
            src_value = getattr(model, attr_name)

            if (isinstance(dst_value, OrderedDict)) or (isinstance(dst_value,
                                                                   List)):
                if len(dst_value) == 0:
                    _to_be_cloned.append([attr_name, src_value])

            elif dst_value is None:
                _to_be_cloned.append([attr_name, src_value])

        for attr_name, src_value in _to_be_cloned:
            setattr(self, attr_name, src_value)

    def dump(self) -> Dict:


        data = {
            "src": "",
            "name": self.name,
            "inputs": {x.name: x.dump() for x in self.inputs},
            "outputs": {x.name: x.dump() for x in self.outputs},
            "format": self.format.name
        }
        if isinstance(self.src, str):
            data["src"] = self.src
        elif isinstance(self.src, Path):
            data["src"] = str(self.src)
        else:
            data["src"] = get_class_name(self.src)

        return data

    def load(self, obj: Dict):
        raise NotImplementedError()



from __future__ import annotations
import inspect
import os
import zipfile
from enum import Enum
from typing import List
from pathlib import Path
import numpy as np

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .dataset import Dataset

from .dataset import DatasetFormat


def _check_object_type(expect_type_str, obj):
    type_str = ""
    try:
        type_str += str(inspect.getmro((obj)))
    except:
        type_str += ""
    type_str += "|"
    type_str += str(type(obj))
    return expect_type_str in type_str


class IdentifierRegistry(type):

    REGISTRY: List = []

    def __new__(cls, name, bases, attrs):
        new_cls = type.__new__(cls, name, bases, attrs)
        cls.REGISTRY.append(new_cls)
        return new_cls


class Identifier(metaclass=IdentifierRegistry):

    FORMAT: DatasetFormat = DatasetFormat.NON_SPECIFIED

    @classmethod
    def is_me(self, dataset: Dataset) -> bool:
        pass


def identifier_selector(dataset: Dataset) -> Identifier:
    print(Identifier.REGISTRY)
    for identifier in Identifier.REGISTRY:
        if identifier.is_me(dataset):
            return identifier

    raise NotImplementedError(f"Unalble to identify {dataset.src}")


def identify(dataset: Dataset) -> Dataset:
    for identifier in Identifier.REGISTRY:
        if identifier.is_me(dataset):
            return identifier.FORMAT

    raise NotImplementedError(f"Unalble to identify {dataset.src}")


class PB(Identifier):
    """
    Use file extension and magic number to identify the file
    """
    FORMAT = DatasetFormat.PB

    @classmethod
    def is_me(cls, dataset):
        if dataset.format == cls.FORMAT:
            return True

        path = dataset.src

        if not (isinstance(path, Path) and path.exists() and path.is_file()):
            return False

        if str(path).lower().endswith('.pb'):
            return True


class NPY(Identifier):
    """
    Use file extension and magic number to identify the file
    """
    FORMAT = DatasetFormat.NPY

    @classmethod
    def is_me(cls, dataset):
        if dataset.format == cls.FORMAT:
            return True

        path = dataset.src

        if not (isinstance(path, Path) and path.exists() and path.is_file()):
            return False

        if str(path).lower().endswith('.npy'):
            return True

        with open(path, 'rb') as f:
            return f.read(6) == bytes.fromhex('934E554D5059')


class NPYDIR(Identifier):
    """
    Use file extension and magic number to identify the file
    """
    FORMAT = DatasetFormat.NPYDIR

    @classmethod
    def is_me(cls, dataset):
        if dataset.format == cls.FORMAT:
            return True

        path = dataset.src

        if not (isinstance(path, Path) and path.exists() and path.is_dir()):
            return False

        files = os.listdir(path)
        if not files:
            return False
        return str(files[0]).lower().endswith('.npy')


class NPZ(Identifier):
    """
    Use file extension and magic number to identify the file
    """
    FORMAT = DatasetFormat.NPZ

    @classmethod
    def is_me(cls, dataset):
        if dataset.format == cls.FORMAT:
            return True

        path = dataset.src

        if not (isinstance(path, Path) and path.exists() and path.is_file()):
            return False

        if str(path).lower().endswith('.npy'):
            return True

        try:
            with zipfile.ZipFile(str(path)) as archive:
                for name in archive.namelist():
                    if not name.endswith('.npy'):
                        continue

                    npy = archive.open(name)
                    version = np.lib.format.read_magic(npy)
                    return True
        except:
            pass

        return False


class NP_ndarray(Identifier):
    """
    Use file extension and magic number to identify the file
    """
    FORMAT = DatasetFormat.NDARRAY

    @classmethod
    def is_me(cls, dataset):
        if dataset.format == cls.FORMAT:
            return True

        return type(dataset.src) is np.ndarray


class NP_memmap(Identifier):
    """
    Use file extension and magic number to identify the file
    """
    FORMAT = DatasetFormat.NP_MEMMAP

    @classmethod
    def is_me(cls, dataset):
        if dataset.format == cls.FORMAT:
            return True

        return type(dataset.src) is np.memmap


class NPZFileObject(Identifier):
    """
    Use file extension and magic number to identify the file
    """
    FORMAT = DatasetFormat.NPZ_OBJECT

    @classmethod
    def is_me(cls, dataset):
        if dataset.format == cls.FORMAT:
            return True

        return dataset.src.__class__.__name__ == "NpzFile"


class TorchDatasetObject(Identifier):
    """
    Use file extension and magic number to identify the file
    """
    FORMAT = DatasetFormat.TORCH_DATASET

    @classmethod
    def is_me(cls, dataset):
        if dataset.format == cls.FORMAT:
            return True

        # from torch.utils.data.dataset import Dataset as TorchDataset
        # return isinstance(dataset.src, TorchDataset)
        str_type = str(dataset.src.__class__.__mro__)
        return '.Dataset' in str_type and 'torch.' in str_type


class TorchDataloaderObject(Identifier):
    """
    Use file extension and magic number to identify the file
    """
    FORMAT = DatasetFormat.TORCH_DATALOADER

    @classmethod
    def is_me(cls, dataset):
        if dataset.format == cls.FORMAT:
            return True

        # from torch.utils.data.dataloader import DataLoader
        # return isinstance(dataset.src, DataLoader)
        str_type = str(dataset.src.__class__.__mro__)
        return '.DataLoader' in str_type and 'torch.' in str_type


class KerasDatasetObject(Identifier):
    """
    Use file extension and magic number to identify the file
    """
    FORMAT = DatasetFormat.KERAS_DATASET

    @classmethod
    def is_me(cls, dataset):
        if dataset.format == cls.FORMAT:
            return True

        ds = dataset.src
        # Keras dataset is a 2x2 metric of tuple
        # ds[0] is for training , ds[0][0] is x and ds[0][1] is y
        # ds[0] is for testing  , ds[1][0] is x and ds[1][1] is y
        # each record is a np.ndarray object

        if len(ds) == 2 and len(ds[0]) == 2 and len(ds[1]) == 2:
            if all([isinstance(x, np.ndarray) for x in [ds[0][0], ds[0][1],
                                                        ds[1][0], ds[1][1]]]):
                return True
        return False


class TFDSPrefetchObject(Identifier):
    """
    Use file extension and magic number to identify the file
    """
    FORMAT = DatasetFormat.TFDS_PREFETCH

    @classmethod
    def is_me(cls, dataset):
        if dataset.format == cls.FORMAT:
            return True
        str_type = str(type(dataset.src))
        return '.PrefetchDataset' in str_type and 'tensorflow.' in str_type



import os
import tempfile


def get_class_name(model):
    c = model.__class__.__mro__[0]
    name = c.__module__ + "." + c.__name__
    return name


def get_tmp_path():
    return os.path.join(tempfile._get_default_tempdir(),
                        next(tempfile._get_candidate_names()))


def remove_if_exist(path):
    if type(path) != str:
        return
    if os.path.exists(path):
        os.remove(path)



import os
from os.path import expanduser
import requests

try:
    if int(os.environ['ONNC_TESTMODE']):
        api_host = 'http://127.0.0.1:8000'
except KeyError:
    api_host = 'https://api.onnc.skymizer.com'

onnc_key_var = 'ONNC_APIKEY'


api_protocol = 'https'
api_url = "api.onnc.skymizer.com"
api_port = 443


from pathlib import Path
from typing import Dict, Union, List, Tuple
import os
import json

from loguru import logger

from .core.modelpackage import ModelPackage
from .core.dataset.dataset import Dataset


from .core.model import ModelFormat, TF_MODEL_FORMATS, ModelDataType
from .core.model.model import Tensor

from .core.dataset import DatasetFormat
from .core.dataset.layout import DataLayout
from .core.project import ProjectData
from .core.compiler.onnc_saas import ONNCSaaSBuilder
from .config import api_protocol, api_url, api_port
from .core.common import get_tmp_path


def login(user: str, password: str):
    os.environ['ONNC_API_USER'] = user
    os.environ['ONNC_API_PASSWD'] = password


def get_api_key():
    return (os.environ.get('ONNC_API_USER'), os.environ.get('ONNC_API_PASSWD'))


class Project:
    def __init__(self, name, builder=ONNCSaaSBuilder,
                 builder_params=[api_protocol, api_url, api_port]):
        self.builder = builder(*builder_params)
        if isinstance(self.builder, ONNCSaaSBuilder):
            self.builder.saas_login(*get_api_key())
            proj_info_sr = self.builder.saas_get_project_id_by_name(name)
            if len(proj_info_sr.data) == 0:
                self.builder.saas_create_project(name)

            logger.info(f"ProjectID: {self.builder._project_id}")

    @classmethod
    def get_model_meta(cls, model_package, inputs_as_nchw):
        if inputs_as_nchw is None:
            if model_package.model.format in TF_MODEL_FORMATS:
                inputs_as_nchw = "as_inputs_name"
            else:
                inputs_as_nchw = False

        model_meta = {
            "inputs": [x.dump() for x in model_package.model.inputs],
            "inputs_as_nchw": inputs_as_nchw,
            "outputs": [x.dump() for x in model_package.model.outputs],
            "format": model_package.model.format.name
        }

        return model_meta

    @classmethod
    def get_dataset_meta(cls, model_package):
        dataset_meta = {
            "shape": model_package.dataset.shape,
            "format": model_package.dataset.format.name
        }

        if isinstance(model_package.dataset.layout, DataLayout):
            dataset_meta["layout"] = model_package.dataset.layout.dump()

        return dataset_meta

    def add_model(self,
                  model: Union[object, str, Path],
                  samples: Union[object, str, Path, Dataset],
                  model_format: ModelFormat = ModelFormat.NON_SPECIFIED,
                  sample_format: DatasetFormat = DatasetFormat.NON_SPECIFIED,
                  # [name: str, shape:Tuple, dtype:ModelDataType]
                  model_inputs: List[Union[List, Tuple, Tensor]] = [],
                  # [name: str, shape:Tuple, dtype:ModelDataType]
                  model_outputs: List[Union[List, Tuple, Tensor]] = [],
                  inputs_as_nchw: Union[str, bool, None, List[str]] = None):
        """Add a model and its corresponding calibration samples.

        :param Union[object, str, Path] model:
            A file or a directory path to a serialized model file,
            or a neurl network model object.

            Check https://docs-tinyonnc.skymizer.com/model-formats.html for
            a list of supported formats.

        :param Union[object, str, Path, Dataset] samples:
            A file or a directory path to a serialized dataset or object.

            Check https://docs-tinyonnc.skymizer.com/dataset-formats.html for
            a list of supported formats.

        :param ModelFormat model_format:
            The format/framework of the model file/object. onnc-bench will
            identify the format if this param is ModelFormat.NON_SPECIFIED.

            Check https://docs-tinyonnc.skymizer.com/model-formats.html for
            a list ofsupported formats.

        :param DatasetFormat sample_format:
            The format/framework of the model file/object. onnc-bench will
            identify the format if this param is DatasetFormat.NON_SPECIFIED.

            Check https://docs-tinyonnc.skymizer.com/dataset-formats.html for
            a list of supported formats.
        :param List[Union[List, Tuple, Tensor]] model_inputs:
            A list of input tensor(s), which include(s) 3 fields:
                1. name: str
                2. shape(optional): Union[List, Tuple]
                3. dtype(optional): onnc.bench.core.model.ModelDataType

        :param List[Union[List, Tuple, Tensor]] model_outputs:
            A list of input tensor(s), which include(s) 3 fields:
                1. name: str
                2. shape(optional): Union[List, Tuple]
                3. dtype(optional): onnc.bench.core.model.ModelDataType


        """
        model_package = ModelPackage(model, samples,
                                     model_format, sample_format,
                                     model_inputs, model_outputs)

        model_tmp_path = Path(get_tmp_path())
        dataset_tmp_path = Path(get_tmp_path())

        model_package.serialize(model_tmp_path, dataset_tmp_path)

        model_meta = self.get_model_meta(model_package, inputs_as_nchw)
        dataset_meta = self.get_dataset_meta(model_package)

        model_id = self.builder.prepare_model(model_package.model.src,
                                              model_package.dataset.src,
                                              model_meta=model_meta,
                                              dataset_meta=dataset_meta)

        self.builder.calibrate(model_id, params={})
        self.builder.compile(model_id, params={})

    def compile(self, device: str, **params):
        """Trigger the compilation process and transform the given
        model into C function calls.

        :param str device:
            The name of the supported SoC board, for example, NUMAKER_IOT_M487.

            Check https://docs-tinyonnc.skymizer.com/devices.html for
            a list of supported devices.

        : param dict params:
            Parameters for the compiler.

            Check https://docs-tinyonnc.skymizer.com/compiler-params.html for
            more details.

        """
        return self.builder.build(device)

    def save(self, path: Union[str, Path]):
        """Save the compiled artifact(s).

        :param Union[str, Path] path:
            A path to store the artifact(s).

        :returns:
            A Deployment object that contains `files`, `report`.
        :rtype:
            Deployment
        """
        if isinstance(path, str):
            path = Path(path)

        return self.builder.save(path)




from typing import Any, List

from abc import abstractmethod
from pathlib import Path
from onnc.bench.core.project import ProjectData
from onnc.bench.core.deployment import Deployment

class IBuilder:

    BUILDER_NAME = ""

    def __init__(self, project_data: ProjectData):
        self.project_data = project_data
        self.model_ids: List = []

    @abstractmethod
    def prepare_model(self, model, dataset, model_meta, dataset_meta) -> Any:
        """Make files of a model and its corresponding dataset ready in given
           path and place.
        """
        pass

    @abstractmethod
    def calibrate(self, model_id, params) -> Any:
        """Calibrate a model package(model and its corresponding samples)
        """
        pass

    @abstractmethod
    def compile(self, model_id, params) -> Any:
        """Compile a model package(model and its corresponding samples)
        """
        pass

    @abstractmethod
    def build(self, device) -> Any:
        """build a project witch contains multiple models

        for model in model_ids:
            ...
        """
        pass

    @abstractmethod
    def save(self, output: Path) -> Deployment:
        pass



from typing import Dict, List, Union, Type, Iterable, Any, Tuple
from pathlib import Path

from loguru import logger

from .model.model import Model, Tensor
from .model.meta import retrive_model_metadata
from .model.transformer import ModelTransformer
from .dataset.dataset import Dataset
from .dataset.transformer import DatasetTransformer
from .evaluator import Evaluator
from .evaluator import Metric
from .model import identifier as model_identifier
from .dataset import identifier as dataset_identifier
from .model import serializer as model_serializer
from .dataset import serializer as dataset_serializer

from .dataset.dataset import Dataset
from .model import ModelFormat, ModelDataType
from .dataset import DatasetFormat
from .common import get_tmp_path


class ModelPackage:
    """Contain all data and controllers

    Data include: model and dataset
    Controllers include: evaluators, model_transformer and dataset_transformer

    """

    def __init__(self, model: Union[object, str, Path, Model],
                 dataset: Union[object, str, Path, Dataset],
                 model_format: ModelFormat = ModelFormat.NON_SPECIFIED,
                 sample_format: DatasetFormat = DatasetFormat.NON_SPECIFIED,
                 #  [name: str, shape:Tuple, dtype:ModelDataType]
                 model_inputs: List[Union[List, Tuple, Tensor]] = [],
                 #  [name: str, shape:Tuple, dtype:ModelDataType]
                 model_outputs: List[Union[List, Tuple, Tensor]] = []):

        self._src_model = model
        self._src_dataset = dataset
        self.evaluators: List[Evaluator] = []
        self.model_transformers: List[ModelTransformer] = []
        self.dataset_transformers: List[DatasetTransformer] = []

        self.__encapsulate_model(model_format, model_inputs,
                                 model_outputs)
        self.__encapsulate_dataset(sample_format)

    def __encapsulate_model(self, model_format, model_inputs, model_outputs):
        # Convert to a Model object if it is not
        if isinstance(self._src_model, Model):
            self.model = self._src_model
        elif isinstance(self._src_model, str):
            self.model = Model(Path(self._src_model))
        else:
            self.model = Model(self._src_model)

        # Retrive inputs
        # if the input name is not specified, retrive automatically
        if len(model_inputs) == 0:
            meta = retrive_model_metadata(self.model)
            for tensor in meta.inputs:
                self.model.add_input(tensor)
        else:
            for idx, i in enumerate(model_inputs):
                if isinstance(i, list) or isinstance(i, tuple):
                    assert len(i) > 0
                    name = i[0]
                    shape = i[1] if len(i) >= 2 else tuple()
                    dtype = i[2] if len(i) >= 3 else ModelDataType.NON_SPECIFIED
                    tensor = Tensor(name, shape, dtype)
                elif isinstance(i, Tensor):
                    tensor = i

                self.model.add_input(tensor)

        # Retrive outputs
        # if the _output name is not specified, retrive automatically
        if len(model_inputs) == 0 and len(model_outputs) == 0:
            for tensor in meta.outputs:
                self.model.add_output(tensor)
        elif len(model_outputs) == 0:
            meta = retrive_model_metadata(self.model)
            for tensor in meta.outputs:
                self.model.add_output(tensor)
        else:
            for idx, i in enumerate(model_outputs):
                if isinstance(i, list) or isinstance(i, tuple):
                    assert len(i) > 0
                    name = i[0]
                    shape = i[1] if len(i) >= 2 else tuple()
                    dtype = i[2] if len(i) >= 3 else ModelDataType.NON_SPECIFIED
                    tensor = Tensor(name, shape, dtype)
                elif isinstance(i, Tensor):
                    tensor = i

                self.model.add_output(tensor)

        # Assign or identify model format
        if not model_format == ModelFormat.NON_SPECIFIED:
            self.model.format = model_format
        else:
            self.model.format = model_identifier.identify(self.model)

        logger.info(self.model.format.name)

    def __encapsulate_dataset(self, dataset_format: DatasetFormat):
        # Convert to a Dataset object if it is not
        if isinstance(self._src_dataset, Dataset):
            self.dataset = self._src_dataset
        elif isinstance(self._src_dataset, str):
            self.dataset = Dataset(Path(self._src_dataset))
        else:
            self.dataset = Dataset(self._src_dataset)

        # specify model format or identify it automatically
        if not dataset_format == DatasetFormat.NON_SPECIFIED:
            self.dataset.format = dataset_format
        else:
            self.dataset.format = dataset_identifier.identify(self.dataset)

    def transform_model(self) -> Type[Model]:
        model = self.model
        for mt in self.model_transformers:
            model = mt.transform(model)
        return model

    def transform_dataset(self) -> Type[Dataset]:
        dataset = self.dataset
        for dt in self.dataset_transformers:
            dataset = dt.transform(dataset)
        return dataset

    def evaluate(self, model, dataset) -> List[Metric]:
        return [e.evaluate(model, dataset) for e in self.evaluators]

    def add_model_transformer(self, model_transformer: ModelTransformer):
        self.model_transformers.append(model_transformer)

    def add_dataset_transformer(self, ds_transformer: DatasetTransformer):
        self.dataset_transformers.append(ds_transformer)

    def add_evaluator(self, evaluator: Evaluator):
        self.evaluators.append(evaluator)

    def serialize(self, model_path: Union[str, Path, None] = None,
                  dataset_path: Union[str, Path, None] = None):

        if model_path is None:
            model_path = get_tmp_path()
        if dataset_path is None:
            dataset_path = get_tmp_path()

        self.model = model_serializer.serialize(self.model, Path(model_path))
        self.dataset = dataset_serializer.serialize(self.dataset,
                                                    Path(dataset_path))
        # we need to get the meta again after serialization
        self.__encapsulate_model(ModelFormat.NON_SPECIFIED, [], [])
        self.__encapsulate_dataset(DatasetFormat.NON_SPECIFIED)

    def dump(self) -> Dict:
        data = {
            "model": self.model.dump(),
            "dataset": self.dataset.dump(),
            "evaluators": [x.dump() for x in self.evaluators],
            "model_transformers": [x.dump() for x in self.model_transformers],
            "dataset_transformers": [x.dump() for x in
                                     self.dataset_transformers]
        }
        return data

    def load(self, data: Dict):
        pass



from __future__ import annotations

from typing import Any, Union, Dict, List, Callable, Tuple, Type
from collections.abc import Iterable
from pathlib import Path
from enum import Enum
from collections import OrderedDict

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .layout import DataLayout
    # from .transformer import PreprocessTransformer
    from .transformer import DatasetTransformer

from ..common import get_class_name
from . import DatasetFormat


class Dataset:
    def __init__(self, raw_dataset: Union[Iterable, Iterable[Path], object]):
        self.src = raw_dataset  # Union[Iteraable, Path]
        self.y_label: OrderedDict[str, Union[int, float, str]] = OrderedDict()
        self.layout: Union[None, DataLayout] = None
        self.format: DatasetFormat = DatasetFormat.NON_SPECIFIED
        self.shape: Union[None, Tuple[int, ...]] = None
        # self._transformers: List = []

    # def exec_transformers(self) -> None:
    #     """To Trigger the transform pipeline
    #     """
    #     dataset = self
    #     for t in self._transformers:
    #         dataset = t.transform(dataset)
    #     self.src = dataset.src

    # def add_transformer(self, transformer: DatasetTransformer) -> None:
    #     self._transformers.append(transformer)

    def clone_attributes(self, dataset: Type[Dataset]) -> None:
        """Clone user defined attributes from input

        Iter all user-defined attributes and check if they have values.
        If the attribute of the destination object has value, then skip
        the clone.

        Args:
            dataset (Type[Dataset]): The dataset object to be cloned
        """
        _to_be_cloned: List = []
        for attr_name in dataset.__dict__:

            dst_value = getattr(self, attr_name)
            src_value = getattr(dataset, attr_name)

            if (isinstance(dst_value, OrderedDict)) or (isinstance(dst_value, List)):
                if len(dst_value) == 0:
                    _to_be_cloned.append([attr_name, src_value])

            elif dst_value is None:
                _to_be_cloned.append([attr_name, src_value])

        for attr_name, src_value in _to_be_cloned:
            setattr(self, attr_name, src_value)

    # def preprocess(self) -> None:
    #     """Alias for exec_transformers
    #     """
    #     self.exec_transformers()

    # def add_preprocessor(self, func: Callable, args: List = [],
    #                      kwargs: Dict = {}) -> None:
    #     """Add a preprocessor in the the pipeline

    #     Several preprocessors can be add into the pipeline

    #     Args:
    #         func (Callable): preprocessor function, the output of the this
    #                          function has to be the preprocessed dataset.
    #         args (List): args for the preprocessor function, use `__DATASET__`
    #                      to indicate the dataset argument position

    #         kwargs (Dict): keword args for the preprocessor function

    #     Example:

    #         def _preprocess(dataset, min_val, max_val, enhancement=False):
    #           ....
    #           ....

    #         add_preprocessor(_preprocess,
    #                         ['__DATASET__', 1, 100],
    #                         {'enhancement': True}
    #                         )
    #     """

    #     pt = PreprocessTransformer()
    #     pt.set_preprocessor(func, args, kwargs)
    #     self.add_transformer(pt)
    #     pass

    def dump(self) -> Dict:
        data = {
            "src": "",
            "y_label": self.y_label,
            "shape": self.shape,
            "layout": self.layout.dump() if self.layout is not None else None,
            "format": self.format.name
        }

        if isinstance(self.src, str):
            data["src"] = self.src
        elif isinstance(self.src, Path):
            data["src"] = str(self.src)
        else:
            data["src"] = get_class_name(self.src)

        return data

    def load(self, obj: Dict):
        raise NotImplementedError()


class ONNCDataset(Dataset):
    def __init__(self):
        super().__init__([])
        self.format = DatasetFormat.ONNC_DATASET


class NDARRAYDataset(Dataset):
    def __init__(self, raw_dataset: Union[Iterable]):
        super().__init__(raw_dataset)
        self.format = DatasetFormat.NDARRAY



from typing import List
from pathlib import Path
from abc import abstractmethod
import shutil
from dataclasses import dataclass

from loguru import logger

from .transformer import ModelTransformer
from .model import Model, Tensor
from . import ModelFormat
from .identifier import identify
from ..common import get_tmp_path


class MetadataRetriverRegistry(type):

    REGISTRY: List = []

    def __new__(cls, name, bases, attrs):
        new_cls = type.__new__(cls, name, bases, attrs)
        cls.REGISTRY.append(new_cls)
        return new_cls


class ModelMeta:
    def __init__(self, inputs: List[Tensor], outputs: List[Tensor]):
        self.inputs = inputs
        self.outputs = outputs


class MetadataRetriver(metaclass=MetadataRetriverRegistry):

    FORMAT: ModelFormat = ModelFormat.NON_SPECIFIED

    @classmethod
    def is_me(cls, model: Model) -> bool:
        return identify(model) == cls.FORMAT

    @abstractmethod
    def retrive_inputs(self, model: Model) -> List[Tensor]:
        raise NotImplementedError("`retrive_inputs` has to be implemented")

    @abstractmethod
    def retrive_outputs(self, model: Model) -> List[Tensor]:
        raise NotImplementedError("`retrive_outputs` has to be implemented")

    def retrive(self, model: Model) -> ModelMeta:
        inputs = self.retrive_inputs(model)
        logger.info(f"Inputs of {model.name}: {inputs}")
        outputs = self.retrive_outputs(model)
        logger.info(f"Outputs of {model.name}: {outputs}")
        return ModelMeta(inputs=self.retrive_inputs(model),
                         outputs=self.retrive_outputs(model))


def metadataretriver_selector(model: Model) -> MetadataRetriver:
    for metadataretriver in MetadataRetriver.REGISTRY:
        if metadataretriver.is_me(model):
            return metadataretriver
    raise NotImplementedError(f"Unalble to retrive metadata from {model}")


def retrive_model_metadata(model: Model) -> ModelMeta:
    for metadataretriver in MetadataRetriver.REGISTRY:
        if metadataretriver.is_me(model):
            return metadataretriver().retrive(model)

    raise NotImplementedError(f"Unalble to retrive metadata from {model}")


# class FileSerializer(MetadataRetriver):

#     FORMAT: ModelFormat = ModelFormat.NON_SPECIFIED

#     def transform(self, model: Model):
#         dest = self.get_param('dest')

#         if model.src.is_file():
#             shutil.move(model.src, dest)

#         elif model.src.is_dir():
#             # Zip the dir if the model is in dir form
#             tmp_path = Path(get_tmp_path())
#             shutil.make_archive(tmp_path, 'zip', model.src)
#             shutil.move(tmp_path / '.zip', dest)

#         return Model(dest)


class H5(MetadataRetriver):

    FORMAT = ModelFormat.H5

    def retrive_inputs(self, model: Model) -> List[Tensor]:
        from tensorflow import keras  # type: ignore[import]
        _model = keras.models.load_model(model.src)

        return [Tensor(name=x.name, shape=tuple(x.shape.as_list()),
                       dtype=x.dtype) for x in _model.inputs]

    def retrive_outputs(self, model: Model) -> List[Tensor]:
        # from tensorflow import keras
        # model = keras.models.load_model(model.src)
        # outputs = {}
        # for x in model.outputs:
        #     inputs[x.name] = Tensor(name=x.name,
        #                             shape=tuple(x.shape.as_list()),
        #                             dtype=x.dtype)
        from tensorflow import keras  # type: ignore[import]
        _model = keras.models.load_model(model.src)

        return [Tensor(name=x.name, shape=tuple(x.shape.as_list()),
                       dtype=x.dtype) for x in _model.outputs]


# class ONNX(MetadataRetriver):

#     FORMAT = ModelFormat.ONNX


# class PTH(MetadataRetriver):

#     FORMAT = ModelFormat.PTH


class PB(MetadataRetriver):

    FORMAT = ModelFormat.PB

    # https://stackoverflow.com/questions/43517959/given-a-tensor-flow-model-graph-how-to-find-the-input-node-and-output-node-name
    def _load_graph(self, frozen_graph_filename):
        import tensorflow as tf
        with tf.io.gfile.GFile(frozen_graph_filename, "rb") as f:
            graph_def = tf.compat.v1.GraphDef()
            graph_def.ParseFromString(f.read())
        with tf.Graph().as_default() as graph:
            tf.import_graph_def(graph_def)

        ops = graph.get_operations()
        return ops

    def analyze_inputs_outputs(graph):

        inputs = []
        for op in ops:
            if len(op.inputs) == 0 and op.type != 'Const':
                inputs.append(op)
            else:
                for input_tensor in op.inputs:
                    if input_tensor.op in outputs_set:
                        outputs_set.remove(input_tensor.op)
        outputs = list(outputs_set)
        return (inputs, outputs)

    def retrive_inputs(self, model: Model) -> List[Tensor]:
        import tensorflow as tf
        ops = self._load_graph(model.src)
        inputs = []
        for op in ops:
            if len(op.inputs) == 0 and op.type != 'Const':
                inputs.append(op)

        return [Tensor(name=x.name, shape=tuple(x.outputs[0].shape.as_list()),
                       dtype=x.outputs[0].dtype) for x in inputs]
        """
        >>> [x.size for x in graph_def.node[0].attr['shape'].shape.dim]
        [-1, 96, 96, 3]
        """
    def retrive_outputs(self, model: Model) -> List[Tensor]:
        import tensorflow as tf
        ops = self._load_graph(model.src)
        inputs = []
        outputs_set = set(ops)

        for op in ops:
            if len(op.inputs) == 0 and op.type != 'Const':
                inputs.append(op)
            else:
                for input_tensor in op.inputs:
                    if input_tensor.op in outputs_set:
                        outputs_set.remove(input_tensor.op)

        outputs = list(outputs_set)

        return [Tensor(name=x.name, shape=tuple(x.outputs[0].shape.as_list()),
                       dtype=x.outputs[0].dtype) for x in outputs]


class SavedModel(MetadataRetriver):

    FORMAT = ModelFormat.SAVED_MODEL

    def retrive_inputs(self, model: Model) -> List[Tensor]:
        from tensorflow import keras  # type: ignore[import]
        _model = keras.models.load_model(model.src)

        return [Tensor(name=x.name, shape=tuple(x.shape.as_list()),
                       dtype=x.dtype) for x in _model.inputs]

    def retrive_outputs(self, model: Model) -> List[Tensor]:
        # from tensorflow import keras
        # model = keras.models.load_model(model.src)
        # outputs = {}
        # for x in model.outputs:
        #     inputs[x.name] = Tensor(name=x.name,
        #                             shape=tuple(x.shape.as_list()),
        #                             dtype=x.dtype)
        from tensorflow import keras  # type: ignore[import]
        _model = keras.models.load_model(model.src)

        return [Tensor(name=x.name, shape=tuple(x.shape.as_list()),
                       dtype=x.dtype) for x in _model.outputs]


class TFKerasModel(MetadataRetriver):

    FORMAT = ModelFormat.TF_KERAS_MODEL

    def retrive_inputs(self, model: Model) -> List[Tensor]:
        return [Tensor(name=x.name, shape=tuple(x.shape.as_list()),
                       dtype=x.dtype) for x in model.src.inputs]

    def retrive_outputs(self, model: Model) -> List[Tensor]:
        return [Tensor(name=x.name, shape=tuple(x.shape.as_list()),
                       dtype=x.dtype) for x in model.src.outputs]


class KerasModel(MetadataRetriver):
    '''
    Keras 2.5.0 Serializer
    '''

    FORMAT = ModelFormat.KERAS_MODEL

    def retrive_inputs(self, model: Model) -> List[Tensor]:
        return [Tensor(name=x.name, shape=tuple(x.shape.as_list()),
                       dtype=x.dtype) for x in model.src.inputs]   # type: ignore[union-attr]

    def retrive_outputs(self, model: Model) -> List[Tensor]:
        return [Tensor(name=x.name, shape=tuple(x.shape.as_list()),
                       dtype=x.dtype) for x in model.src.outputs]  # type: ignore[union-attr]


class PytorchModel(MetadataRetriver):
    """Use python MRO to check if it contains specific str"""

    FORMAT = ModelFormat.PT_NN_MODULE

    def retrive_inputs(self, model: Model) -> List[Tensor]:
        _model = model.src
        tensor = list(_model.parameters())[  # type: ignore[union-attr]
            0].detach()
        t_name = "input" if tensor.name == None else tensor.name
        tensor = tensor.numpy()
        return [Tensor(t_name, tensor.shape, tensor.dtype)]

    def retrive_outputs(self, model: Model) -> List[Tensor]:
        _model = model.src

        # type: ignore[union-attr]
        tensor = list(_model.parameters())[-1].detach()
        t_name = "output" if tensor.name == None else tensor.name
        tensor = tensor.numpy()
        return [Tensor(t_name, tensor.shape, tensor.dtype)]



from typing import Union, List
from abc import abstractmethod
from pathlib import Path
import shutil
import numpy as np
from loguru import logger

from .identifier import DatasetFormat, identify
from ..common import get_tmp_path
from .transformer import DatasetTransformer
from .dataset import Dataset


class SerializerRegistry(type):

    REGISTRY: List = []

    def __new__(cls, name, bases, attrs):

        # inherent attrs in ModelTransformer but not overwrite them
        for k in DatasetTransformer.__dict__:
            if k not in attrs:
                attrs[k] = DatasetTransformer.__dict__[k]

        new_cls = type.__new__(cls, name, bases, attrs)
        cls.REGISTRY.append(new_cls)
        return new_cls


class Serializer(metaclass=SerializerRegistry):

    FORMAT: Union[None, DatasetFormat] = None

    @classmethod
    def is_me(cls, dataset: Dataset) -> bool:
        return identify(dataset) == cls.FORMAT

    @abstractmethod
    def transform(self, dataset: Dataset) -> Dataset:
        pass

    def serialize(self, dataset: Dataset, dest: Path) -> Dataset:
        self.add_param('dest', dest)  # type: ignore[attr-defined]
        return self.transform(dataset)


def serializer_selector(dataset: Dataset) -> Serializer:
    for serializer in Serializer.REGISTRY:
        if serializer.is_me(dataset):
            return serializer

    raise NotImplementedError(f"Unalble to identify {dataset.src}")


def serialize(dataset: Dataset, dest: Path):
    for serializer in Serializer.REGISTRY:
        if serializer.is_me(dataset):
            return serializer().serialize(dataset, dest)

    raise NotImplementedError(f"Unalble to identify {dataset.src}")


class FileSerializer(Serializer):

    FORMAT = DatasetFormat.NON_SPECIFIED

    def transform(self, dataset: Dataset):
        dest = self.get_param('dest')  # type: ignore[attr-defined]

        if isinstance(dataset.src, Path):

            if dataset.src.is_file():
                shutil.move(str(dataset.src), dest)

            elif dataset.src.is_dir():
                tmp_path = Path(get_tmp_path())
                shutil.make_archive(str(tmp_path), 'zip', dataset.src)
                shutil.move(str(tmp_path / '.zip'), dest)

            return Dataset(dest)

        raise ValueError("FileSerializer accepts only Path-like objects")


class NPY(FileSerializer):

    FORMAT = DatasetFormat.NPY


class NPYDIR(FileSerializer):

    FORMAT = DatasetFormat.NPYDIR


class NPZ(FileSerializer):

    FORMAT = DatasetFormat.NPZ


class NDARRAY(Serializer):

    FORMAT = DatasetFormat.NDARRAY

    def transform(self, dataset: Dataset):
        dest = self.get_param('dest')  # type: ignore[attr-defined]
        np.save(dest, dataset.src)
        shutil.move(dest.with_suffix('.npy'), dest)
        return Dataset(dest)


class TORCH_DATASET(Serializer):

    FORMAT = DatasetFormat.TORCH_DATASET

    def transform(self, dataset: Dataset):
        dest = self.get_param('dest')  # type: ignore[attr-defined]
        dataset_npy = Dataset(dataset.src.data.numpy()) # type: ignore[union-attr]
        ndarray_serializer = NDARRAY()
        return ndarray_serializer.serialize(dataset_npy, dest)


class TORCH_DATALOADER(Serializer):

    FORMAT = DatasetFormat.TORCH_DATALOADER

    def transform(self, dataset: Dataset):
        dest = self.get_param('dest')  # type: ignore[attr-defined]
        dataset_torch = Dataset(dataset.src.dataset) # type: ignore[union-attr]
        torchdataset_serializer = TORCH_DATASET()
        return torchdataset_serializer.serialize(dataset_torch, dest)


class KERAS_DATASET(Serializer):

    FORMAT = DatasetFormat.KERAS_DATASET

    def transform(self, dataset: Dataset):
        dest = self.get_param('dest')  # type: ignore[attr-defined]
        # Keras dataset is a 2x2 metric of tuple
        # ds[0] is for training , ds[0][0] is x and ds[0][1] is y
        # ds[0] is for testing  , ds[1][0] is x and ds[1][1] is y
        # each record is a np.ndarray object
        # hence, we use x of testing data here
        dataset_npy = Dataset(dataset.src[1][0])  #type: ignore[index]
        ndarray_serializer = NDARRAY()
        return ndarray_serializer.serialize(dataset_npy, dest)


class TFDS_PREFETCH(Serializer):

    FORMAT = DatasetFormat.TFDS_PREFETCH

    def transform(self, dataset: Dataset):
        LABEL_KEYS = ['label']

        import tensorflow_datasets as tfds  # type: ignore[import]

        dest = self.get_param('dest')  # type: ignore[attr-defined]
        try:
            sample_num = self.get_param('sample_num') #type: ignore[attr-defined]

            if not sample_num:
                logger.warning('Parameter `sample_num` is not set, '
                               'use default value `1000`')
                sample_num = 1000
        except Exception as e:
            logger.warning(
                'Parameter `sample_num` is not set, use default value `1000`')
            sample_num = 1000

        ds = [x for x in dataset.src.shuffle(  # type: ignore[union-attr]
              1024).take(1000).as_numpy_iterator()]

        keys = list(ds[0].keys())
        if len(keys) > 1:
            try:
                for key in LABEL_KEYS:
                    keys.remove(key)
            except ValueError:
                pass

        if len(keys) > 1:
            try:
                data_key = self.get_param('data_key') #type: ignore[attr-defined]
                if not data_key:
                    logger.warning(f'Parameter `data_key` is not set, '
                                   'use default value `{keys[0]}`')
                    data_key = keys[0]
            except Exception as e:
                logger.warning(f'Parameter `data_key` is not set, '
                               'use default value `{keys[0]}`')
                data_key = keys[0]
        else:
            data_key = keys[0]

        dataset_npy = Dataset(np.stack([x[data_key] for x in ds]))

        ndarray_serializer = NDARRAY()
        return ndarray_serializer.serialize(dataset_npy, dest)



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Dict
from .project import ProjectData
from .deployment import Deployment
from .benchmark import Metric
from .compiler import ONNCSaaSCompiler
from ..config import api_protocol, api_url, api_port
from .model.serializer import serialize as serialize_model
from .dataset.serializer import serialize as serialize_dataset
from ..utils import get_tmp_path


class IBuilder(ABC):
    def __init__(self, name: str, info: Dict) -> None:
        self.project_data = None

    @abstractmethod
    def calibrate(self, model: object, dataset: object, parameters: Dict):
        pass

    @abstractmethod
    def build(self, device: str, parameters: Dict):
        pass

    @abstractmethod
    def save(self, output: Path) -> Deployment:
        pass

    @abstractmethod
    def measure_perf(self) -> Metric:
        pass


class ONNCSaaSAPISBuilder(IBuilder):
    def __init__(self, project: ProjectData):
        self.project = project
        self.compiler = ONNCSaaSCompiler(protocol=api_protocol,
                                         url=api_url, port=api_port)
        self._to_be_removed = []

        def __init__(name: str, info: Dict) -> None:
            pass

        def calibrate(self, parameters):
            for mp in self.project.model_packages:
                model = mp.model
                dataset = mp.dataset

                tmp_model_path = get_tmp_path()
                tmp_ds_path = get_tmp_path()

                self._to_be_removed.append(tmp_model_path)
                self._to_be_removed.append(tmp_ds_path)

                serialize_model(model, tmp_model_path)
                serialize_dataset(dataset, tmp_ds_path)

                self.compiler.upload_model(tmp_model_path, tmp_ds_path)

        def build(self, device: str, params: Dict):
            params['device': device]
            self.compiler.compile(params)

        def save(path: Path):
            self.compiler.download(path)

        def measure_perf(device: str):
            pass




from pathlib import Path
from typing import Dict, Union, List, Tuple
import json
from .modelpackage import ModelPackage
from .dataset.dataset import Dataset
from .model import ModelFormat, ModelDataType

from .dataset import DatasetFormat


class ProjectData:

    def __init__(self):
        self.project_id = None  # str
        self.model_packages = []  # List[ModelPackage]

    def add_model(self, model_package: ModelPackage) -> None:
        self.model_packages.append(model_package)

    def dump(self) -> Dict:
        # TODO
        # check evaluators, preprocessors, trainers is scirpt type

        mp = [m.dump() for m in self.model_packages]

        meta = {
            "project_id": self.project_id,
            "model_packages": mp
        }
        return meta

    def load(self, meta: Dict):
        self.project_id = meta["project_id"]
        self.model_packages = []
        for model_package in meta["model_packages"]:
            mp = ModelPackage(model_package["model"]["src"],
                              model_package["dataset"]["src"])

            mp.load(model_package)
            self.model_packages.append(mp)



from enum import Enum

class ModelFormat(Enum):
    NON_SPECIFIED = 0
    UNKNOWN = 1

    PT_NN_MODULE = 11
    TF_KERAS_MODEL = 12
    KERAS_MODEL = 13

    H5 = 21
    SAVED_MODEL = 22
    PB = 23

    PTH = 31
    ONNX = 32


TF_MODEL_FORMATS = [ModelFormat.TF_KERAS_MODEL,
                    ModelFormat.KERAS_MODEL,
                    ModelFormat.H5,
                    ModelFormat.SAVED_MODEL,
                    ModelFormat.PB]


class ModelDataType(Enum):
    NON_SPECIFIED = 0
    FP32 = 1
    FP16 = 2

    BF16 = 11

    INT8 = 21



from pathlib import Path
from typing import Dict, List, Any, Callable, Union
from abc import abstractmethod
from pathlib import Path
import inspect
import os
import time
import re
import zipfile
from dataclasses import dataclass
import sys

import requests
from loguru import logger

from onnc.bench.core.project import ProjectData
from onnc.bench.core.deployment import Deployment
from .builder import IBuilder
from ..dataset.dataset import Dataset
from ..model.model import Model
from .saas_config import URI_MAP, timeout, poll_interval
from ..modelpackage import ModelPackage
from onnc.bench.core.common import get_tmp_path


def spinning_cursor():
    while True:
        for cursor in '|/-\\':
            yield cursor


class SaaSResult:
    def __init__(self, http_res):
        self.success = None
        self.message = None
        self.http_status_code = None
        self.data: Dict = {}

        self.from_http_res(http_res)

    def __bool__(self):
        return True if self.success else False

    def from_http_res(self, http_res) -> None:
        self.http_status_code = http_res.status_code

        if http_res.status_code == requests.codes.ok:
            self.success = True
            self.message = ""
            try:
                self.data = http_res.json()
            except Exception:
                self.data = {}
        else:
            self.success = False
            self.message = http_res.json()["error"]["message"]
            self.data = http_res.json()


class Compilation:
    def __init__(self):
        self.compilation_id: str  # SaaS compilation_id
        self.model_path: str
        self.sample_path: str
        self.model_meta: Dict[str, Any]
        self.sample_meta: Dict[str, Any]
        self.calibrator_params: Dict[str, Any]
        self.compiler_params: Dict[str, Any]


class ONNCSaaSBuilder(IBuilder):
    BUILDER_NAME = "ONNCSaaSBuilder"

    def __init__(self, protocal: str, url: str, port: int):
        self.base_url = f"{protocal}://{url}:{port}"
        self.auth_token = None
        self._compilations: Dict[int, Compilation] = {}
        self._project_id: str = ""
        self._build_id: str = ""

    def _http_req(self, api: Union[Callable, List], *args, **kwargs):
        if type(api) == list:
            uri = URI_MAP[api[0]]["uri"].format(*api[1:])
            method = URI_MAP[api[0]]["method"]
        elif callable(api):
            uri = URI_MAP[api.__name__]["uri"]
            method = URI_MAP[api.__name__]["method"]

        if method == 'GET':
            return requests.get(f"{self.base_url}{uri}", *args, **kwargs)
        elif method == 'POST':
            return requests.post(f"{self.base_url}{uri}", *args, **kwargs)
        elif method == 'DELETE':
            return requests.delete(f"{self.base_url}{uri}", *args, **kwargs)
        elif method == 'PATCH':
            return requests.patch(f"{self.base_url}{uri}", *args, **kwargs)
        elif method == 'PUT':
            return requests.put(f"{self.base_url}{uri}", *args, **kwargs)

    def saas_login(self, email: str, password: str) -> SaaSResult:
        data = {
            "email": email,
            "password": password
        }
        r = self._http_req(self.saas_login, json=data)
        if r.status_code == requests.codes.ok:
            self.auth_token = r.json()["token"]

        return SaaSResult(r)

    def saas_verify_key(self, api_key) -> bool:
        pass

    def saas_create_project(self, name: str, info: Dict = {}) -> SaaSResult:
        data = {
            "name": name,
            "info": info
        }
        r = self._http_req(self.saas_create_project, json=data)
        if r.status_code == requests.codes.ok:
            self._project_id = r.json()["id"]

        return SaaSResult(r)

    def saas_get_project_id_by_name(self, project_name: str) -> SaaSResult:
        r = self._http_req(['saas_get_project_id_by_name', project_name])
        r_list = r.json()
        if len(r_list) > 0:
            self._project_id = r_list[0]["id"]

        return SaaSResult(r)

    def saas_upload_file(self, file_path: str) -> SaaSResult:
        files = {os.path.basename(file_path): open(file_path, 'rb')}

        r = self._http_req(self.saas_upload_file, files=files)

        return SaaSResult(r)

    def saas_create_compilation(self, model_path, dataset_path,
                                params: Dict = {}):

        model_sr = self.saas_upload_file(model_path)  # type: ignore[arg-type]

        dataset_sr = self.saas_upload_file(dataset_path)  # type: ignore[arg-type]
        if not model_sr:
            return model_sr

        if not dataset_sr:
            return dataset_sr

        data = {
            "model": model_sr.data["files"][0]["url"],
            "modelSize": model_sr.data["files"][0]["size"],
            "calibration": dataset_sr.data["files"][0]["url"],
            "calibrationSize": dataset_sr.data["files"][0]["size"],
            "compilerParameters": params
        }

        r = self._http_req(self.saas_create_compilation, json=data)
        return SaaSResult(r)

    def saas_create_build(self, project_id: str, saas_compilations: List,
                          device: str) -> SaaSResult:
        data = {
            # "userId": "0",
            "projectId": project_id,
            "boardId": device,
            "input": saas_compilations
        }

        r = self._http_req(self.saas_create_build, json=data)

        return SaaSResult(r)

    def saas_get_build_state(self, build_id: str) -> SaaSResult:
        """Get state code of a build.

        :returns:
            The state code of the build
        :rtype:
            str
        """
        r = self._http_req(['saas_get_build_state', build_id])

        return SaaSResult(r)

    def saas_download_deployment_package(self, file_id, dest: Path):
        r = self._http_req(['saas_download_deployment_package', file_id],
                           stream=True)

        r.raise_for_status()

        with open(dest, 'wb') as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)

        return SaaSResult(r)

    def saas_upload_report(self, compilation_id: str, report_type: str,
                           report: Dict) -> SaaSResult:
        data = {
            "values": report,
            "compilationId": compilation_id,
            "report_type": report_type
        }

        r = self._http_req(self.saas_upload_report, json=data)

        return SaaSResult(r)

    def saas_list_devices(self) -> SaaSResult:
        r = self._http_req(self.saas_list_devices)
        return SaaSResult(r)

    def prepare_model(self, model_path, dataset_path,
                      model_meta, dataset_meta) -> int:
        """ Upload a model and its corresponding calibration dataset.
            And create a compilation.

        :param str model:
            A path to a model file
        :param str dataset:
            A path to a model file
        :param Dict model_meta
            Metadata of the model
        :param Dict dataset_meta
            Metadata of the dataset
        :returns:
            internal compilation id
        :rtype:
            int
        """

        compilation = Compilation()
        compilation.model_path = model_path
        compilation.sample_path = dataset_path
        compilation.model_meta = model_meta
        compilation.sample_meta = dataset_meta

        _internal_cid = id(model_path)
        self._compilations[_internal_cid] = compilation
        return _internal_cid

    def calibrate(self, model_id: int, params: Dict):
        """Update calibration parameters
        """
        self._compilations[model_id].calibrator_params = params

    def compile(self, model_id, params: Dict):
        """Update compilation parameters
        """
        self._compilations[model_id].compiler_params = params

    def build(self, device: str) -> Dict:

        saas_compilations = []

        # Upload files and create compilation
        for iternal_cid in self._compilations:
            compilation = self._compilations[iternal_cid]

            compilation.compiler_params['device'] = device

            params = {}
            params["model_meta"] = compilation.model_meta
            params["sample_meta"] = compilation.sample_meta
            params["compiler_params"] = compilation.compiler_params
            params["calibrator_params"] = compilation.calibrator_params

            cr = self.saas_create_compilation(compilation.model_path,
                                              compilation.sample_path,
                                              params)
            # saas_compilations.append(cr.data["id"])
            saas_compilations.append(cr.data)

        # Create Build and trigger build/compilation
        build_sr = self.saas_create_build(project_id=self._project_id,
                                          saas_compilations=saas_compilations,
                                          device=device)
        if not build_sr:
            return build_sr.data

        self._build_id = build_sr.data["id"]

        logger.info(f"ONNC-SAAS BuildID: {self._build_id}")

        # Wait and poll the result
        t = 0
        spinner = spinning_cursor()
        sys.stdout.write("Building... ")
        while t < (timeout / poll_interval):
            t += 1

            # update spinner
            sys.stdout.write(next(spinner))
            sys.stdout.flush()
            sys.stdout.write('\b')
            time.sleep(poll_interval)

            state_sr = self.saas_get_build_state(build_sr.data["id"])
            if not state_sr:
                return state_sr.data
            elif state_sr and (state_sr.data["state"] in ["pending",
                                                          "running"]):
                time.sleep(poll_interval)
            else:
                return state_sr.data

        return {"error": {"message": "Timeout"}}

    def save(self, output: Path) -> Deployment:
        bs = self.saas_get_build_state(self._build_id)
        if bs.data["state"] == "success":
            logger.success(f"Compiled successfully.")
            file_url = bs.data["output"]["deploymentPackage"]

            logger.info(file_url)
            
            file_ids = re.findall('files/(.*?)$', file_url)
            if len(file_ids) != 1:
                logger.error(f"Parse file_id error: {file_url}")
            else:
                file_id = file_ids[0]

            tmp_download = get_tmp_path() + '.zip'
            
            logger.info(tmp_download)
            
            self.saas_download_deployment_package(file_id, tmp_download)

            with zipfile.ZipFile(tmp_download, 'r') as zip_ref:
                zip_ref.extractall(output)

            deployment = Deployment(output)
        else:
            logger.error(f"Compilation failed: {bs.data}")
            deployment = Deployment(Path(""))
        return deployment

    def get_device_list(self) -> Dict:
        response = requests.get(f"{self.base_url}/api/v1/supprted/devices/")
        response = response.json()
        if ("success" in response) and (response["success"]):
            return response["payload"]
        return response.json()



from typing import List, Dict, Union
from pathlib import Path
import shutil
import json
import os


class Deployment:
    """
    Deployment is a container class for built artifacts and reports
    """
    META_FNAME = Path('.deployment.json')

    def __init__(self, path: Union[None, Path]):
        if path:
            self.base_path = Path(path)

            # if self.base_path.joinpath(self.META_FNAME).exists():
            #     self.load()
            # else:
            self.compiled_files = [
                x for x in os.listdir(self.base_path / 'src')]
            self.report_path = self.base_path / 'report.json'
            self.report = dict(json.loads(open(self.report_path, 'r').read()
                                          )["metrics"])
            self.compile_logs = json.loads(open(self.report_path, 'r').read()
                                           )["logs"]

            # self.save()

    def __str__(self):
        return json.dumps(self.meta, sort_keys=True, indent=2)

    def __repr__(self):
        return self.meta

    @property
    def meta(self):
        return {"base_path": str(self.base_path),
                "compiled_files": [str(x) for x in self.compiled_files],
                "report_path": str(self.report_path),
                "report": self.report
                }

    def save(self):
        _path = self.base_path / self.META_FNAME
        open(_path, 'w').write(json.dumps(self.meta, sort_keys=True, indent=4))

    def load(self):
        meta = json.loads(open(self.base_path / self.META_FNAME, 'r').read())

        self.base_path = meta["base_path"]
        self.compiled_files = meta["compiled_files"]
        self.report_path = meta["report_path"]
        self.report = json.loads(open(meta["report_path"], "r").read())

    def load_raw(self):
        """Scan folder and construct the object"""
        pass

    def deploy(self, target: Path):
        """Copy the deployment folder to target

        Copy the deployment folder to target and reconstruct the meta

        """
        os.remove(self.base_path / self.META_FNAME)
        shutil.copytree(self.base_path, target)
        deployment = Deployment(target)
        return deployment
